@@ -1,5 +1,5 @@
 //
-// Copyright (C) 2010-2021 Codership Oy <info@codership.com>
+// Copyright (C) 2010-2014 Codership Oy <info@codership.com>
 //
 
 #include "galera_common.hpp"
@@ -31,7 +31,18 @@ apply_trx_ws(void*                    recv_ctx,
     {
         try
         {
+            if (trx.is_toi())
+            {
+                log_debug << "Executing TO isolated action: " << trx;
+            }
+
             gu_trace(trx.apply(recv_ctx, apply_cb, meta));
+
+            if (trx.is_toi())
+            {
+                log_debug << "Done executing TO isolated action: "
+                         << trx.global_seqno();
+            }
             break;
         }
         catch (galera::ApplyException& e)
@@ -131,10 +142,8 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
                              config_.get(Param::commit_order))),
     state_file_         (config_.get(BASE_DIR)+'/'+GALERA_STATE_FILE),
     st_                 (state_file_),
-    safe_to_bootstrap_  (true),
     trx_params_         (config_.get(BASE_DIR), -1,
                          KeySet::version(config_.get(Param::key_format)),
-                         TrxHandle::Defaults.record_set_ver_,
                          gu::from_string<int>(config_.get(
                              Param::max_write_set_size))),
     uuid_               (WSREP_UUID_UNDEFINED),
@@ -155,7 +164,6 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     sst_mutex_          (),
     sst_cond_           (),
     sst_retry_sec_      (1),
-    last_st_type_       (ST_TYPE_NONE),
     gcache_             (config_, config_.get(BASE_DIR)),
     gcs_                (config_, gcache_, proto_max_, args->proto_ver,
                          args->node_name, args->node_incoming),
@@ -211,8 +219,6 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     state_.add_transition(Transition(S_JOINED, S_CLOSING));
     state_.add_transition(Transition(S_JOINED, S_CONNECTED));
     state_.add_transition(Transition(S_JOINED, S_SYNCED));
-    // the following is possible if one desync() immediately follows another
-    state_.add_transition(Transition(S_JOINED, S_DONOR));
 
     state_.add_transition(Transition(S_SYNCED, S_CLOSING));
     state_.add_transition(Transition(S_SYNCED, S_CONNECTED));
@@ -227,7 +233,7 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     wsrep_uuid_t  uuid;
     wsrep_seqno_t seqno;
 
-    st_.get (uuid, seqno, safe_to_bootstrap_);
+    st_.get (uuid, seqno);
 
     if (0 != args->state_id &&
         args->state_id->uuid != WSREP_UUID_UNDEFINED &&
@@ -242,16 +248,13 @@ galera::ReplicatorSMM::ReplicatorSMM(const struct wsrep_init_args* args)
     log_debug << "End state: " << uuid << ':' << seqno << " #################";
 
     update_state_uuid (uuid);
-    gcache_.seqno_reset(to_gu_uuid(uuid), seqno);
-    // update gcache position to one supplied by app.
 
     cc_seqno_ = seqno; // is it needed here?
-
-    // the following initialization is needed only to pass seqno to
-    // connect() call. Ideally this should be done only on receving conf change.
     apply_monitor_.set_initial_position(seqno);
+
     if (co_mode_ != CommitOrder::BYPASS)
         commit_monitor_.set_initial_position(seqno);
+
     cert_.assign_initial_position(seqno, trx_proto_ver());
 
     build_stats_vars(wsrep_stats_);
@@ -268,7 +271,6 @@ galera::ReplicatorSMM::~ReplicatorSMM()
     case S_SYNCED:
     case S_DONOR:
         close();
-        // fall through
     case S_CLOSING:
         // @todo wait that all users have left the building
     case S_CLOSED:
@@ -290,29 +292,19 @@ wsrep_status_t galera::ReplicatorSMM::connect(const std::string& cluster_name,
 
     ssize_t err;
     wsrep_status_t ret(WSREP_OK);
-    wsrep_seqno_t const seqno(STATE_SEQNO());
+    wsrep_seqno_t const seqno(cert_.position());
     wsrep_uuid_t  const gcs_uuid(seqno < 0 ? WSREP_UUID_UNDEFINED :state_uuid_);
 
     log_info << "Setting initial position to " << gcs_uuid << ':' << seqno;
 
-    if ((bootstrap == true || cluster_url == "gcomm://")
-        && safe_to_bootstrap_ == false)
-    {
-        log_error << "It may not be safe to bootstrap the cluster from this node. "
-                  << "It was not the last one to leave the cluster and may "
-                  << "not contain all the updates. To force cluster bootstrap "
-                  << "with this node, edit the grastate.dat file manually and "
-                  << "set safe_to_bootstrap to 1 .";
-        ret = WSREP_NODE_FAIL;
-    }
-
-    if (ret == WSREP_OK &&
-        (err = gcs_.set_initial_position(gcs_uuid, seqno)) != 0)
+    if ((err = gcs_.set_initial_position(gcs_uuid, seqno)) != 0)
     {
         log_error << "gcs init failed:" << strerror(-err);
         ret = WSREP_NODE_FAIL;
     }
 
+    gcache_.reset();
+
     if (ret == WSREP_OK &&
         (err = gcs_.connect(cluster_name, cluster_url, bootstrap)) != 0)
     {
@@ -342,6 +334,8 @@ wsrep_status_t galera::ReplicatorSMM::close()
 
 wsrep_status_t galera::ReplicatorSMM::async_recv(void* recv_ctx)
 {
+    assert(recv_ctx != 0);
+
     if (state_() == S_CLOSED || state_() == S_CLOSING)
     {
         log_error <<"async recv cannot start, provider in closed/closing state";
@@ -356,8 +350,6 @@ wsrep_status_t galera::ReplicatorSMM::async_recv(void* recv_ctx)
 
     while (WSREP_OK == retval && state_() != S_CLOSING)
     {
-        GU_DBUG_SYNC_EXECUTE("before_async_recv_process_sync", sleep(5););
-
         ssize_t rc;
 
         while (gu_unlikely((rc = as_->process(recv_ctx, exit_loop))
@@ -371,15 +363,7 @@ wsrep_status_t galera::ReplicatorSMM::async_recv(void* recv_ctx)
 
         if (gu_unlikely(rc <= 0))
         {
-            if (GcsActionSource::INCONSISTENCY_CODE == rc)
-            {
-                st_.mark_corrupt();
-                retval = WSREP_FATAL;
-            }
-            else
-            {
-                retval = WSREP_CONN_FAIL;
-            }
+            retval = WSREP_CONN_FAIL;
         }
         else if (gu_unlikely(exit_loop == true))
         {
@@ -446,12 +430,6 @@ void galera::ReplicatorSMM::apply_trx(void* recv_ctx, TrxHandle* trx)
     wsrep_trx_meta_t meta = {{state_uuid_, trx->global_seqno() },
                              trx->depends_seqno()};
 
-    if (trx->is_toi())
-    {
-        log_debug << "Executing TO isolated action: " << *trx;
-        st_.mark_unsafe();
-    }
-
     gu_trace(apply_trx_ws(recv_ctx, apply_cb_, commit_cb_, *trx, meta));
     /* at this point any exception in apply_trx_ws() is fatal, not
      * catching anything. */
@@ -494,13 +472,6 @@ void galera::ReplicatorSMM::apply_trx(void* recv_ctx, TrxHandle* trx)
 
     apply_monitor_.leave(ao);
 
-    if (trx->is_toi())
-    {
-        log_debug << "Done executing TO isolated action: "
-                  << trx->global_seqno();
-        st_.mark_safe();
-    }
-
     trx->set_exit_loop(exit_loop);
 }
 
@@ -870,8 +841,8 @@ wsrep_status_t galera::ReplicatorSMM::replay_trx(TrxHandle* trx, void* trx_ctx)
         ApplyOrder ao(*trx);
         gu_trace(apply_monitor_.enter(ao));
         trx->set_state(TrxHandle::S_MUST_REPLAY_CM);
+        // fall through
     }
-    // fall through
     case TrxHandle::S_MUST_REPLAY_CM:
         if (co_mode_ != CommitOrder::BYPASS)
         {
@@ -978,19 +949,12 @@ wsrep_status_t galera::ReplicatorSMM::post_rollback(TrxHandle* trx)
 
 wsrep_status_t galera::ReplicatorSMM::causal_read(wsrep_gtid_t* gtid)
 {
-    wsrep_seqno_t cseq;
-    gu::datetime::Date wait_until(gu::datetime::Date::calendar() +
-                                  causal_read_timeout_);
+    wsrep_seqno_t cseq(static_cast<wsrep_seqno_t>(gcs_.caused()));
 
-    try
+    if (cseq < 0)
     {
-        gcs_.caused(cseq, wait_until);
-        assert(cseq >= 0);
-    }
-    catch (gu::Exception& e)
-    {
-        log_warn << "gcs_caused() returned " << -e.get_errno()
-                 << " (" << strerror(e.get_errno()) << ")";
+        log_warn << "gcs_caused() returned " << cseq << " (" << strerror(-cseq)
+                 << ')';
         return WSREP_TRX_FAIL;
     }
 
@@ -1003,6 +967,8 @@ wsrep_status_t galera::ReplicatorSMM::causal_read(wsrep_gtid_t* gtid)
         // at monitor drain and disallowing further waits until
         // configuration change related operations (SST etc) have been
         // finished.
+        gu::datetime::Date wait_until(gu::datetime::Date::calendar()
+                                      + causal_read_timeout_);
         if (gu_likely(co_mode_ != CommitOrder::BYPASS))
         {
             commit_monitor_.wait(cseq, wait_until);
@@ -1118,9 +1084,8 @@ writeset_from_handle (wsrep_po_handle_t& handle,
 //                gu::String<256>(trx_params.working_dir_) << '/' << &handle,
                 trx_params.working_dir_, wsrep_trx_id_t(&handle),
                 /* key format is not essential since we're not adding keys */
-                KeySet::version(trx_params.key_format_), NULL, 0, 0,
-                trx_params.record_set_ver_,
-                WriteSetNG::MAX_VERSION, DataSet::MAX_VERSION, DataSet::MAX_VERSION,
+                KeySet::version(trx_params.key_format_), NULL, 0,
+                0, WriteSetNG::MAX_VERSION, DataSet::MAX_VERSION, DataSet::MAX_VERSION,
                 trx_params.max_write_set_size_);
 
             handle.opaque = ret;
@@ -1264,7 +1229,7 @@ void galera::ReplicatorSMM::process_trx(void* recv_ctx, TrxHandle* trx)
 
             log_fatal << "Failed to apply trx: " << *trx;
             log_fatal << e.what();
-            log_fatal << "Node consistency compromised, aborting...";
+            log_fatal << "Node consistency compromized, aborting...";
             abort();
         }
         break;
@@ -1301,8 +1266,6 @@ void galera::ReplicatorSMM::process_commit_cut(wsrep_seqno_t seq,
 
 void galera::ReplicatorSMM::establish_protocol_versions (int proto_ver)
 {
-    trx_params_.record_set_ver_ = gu::RecordSet::VER1;
-
     switch (proto_ver)
     {
     case 1:
@@ -1333,18 +1296,6 @@ void galera::ReplicatorSMM::establish_protocol_versions (int proto_ver)
         trx_params_.version_ = 3;
         str_proto_ver_ = 2;
         break;
-    case 8:
-        // Protocol upgrade to enforce 8-byte alignment in writesets.
-        trx_params_.version_ = 3;
-        trx_params_.record_set_ver_ = gu::RecordSet::VER2;
-        str_proto_ver_ = 2;
-        break;
-    case 9:
-        // Protocol upgrade to enable support for semi-shared key type.
-        trx_params_.version_ = 4;
-        trx_params_.record_set_ver_ = gu::RecordSet::VER2;
-        str_proto_ver_ = 2;
-        break;
     default:
         log_fatal << "Configuration change resulted in an unsupported protocol "
             "version: " << proto_ver << ". Can't continue.";
@@ -1403,7 +1354,6 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
                                            State                    next_state,
                                            wsrep_seqno_t            seqno_l)
 {
-    assert(repl_proto >= 0 || view_info.status != WSREP_VIEW_PRIMARY);
     assert(seqno_l > -1);
 
     update_incoming_list(view_info);
@@ -1413,11 +1363,6 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
 
     wsrep_seqno_t const upto(cert_.position());
 
-    if (view_info.status == WSREP_VIEW_PRIMARY)
-    {
-        safe_to_bootstrap_ = (view_info.memb_num == 1);
-    }
-
     apply_monitor_.drain(upto);
 
     if (co_mode_ != CommitOrder::BYPASS) commit_monitor_.drain(upto);
@@ -1440,9 +1385,6 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
         if (S_CONNECTED != state_()) state_.shift_to(S_CONNECTED);
     }
 
-    // must establish protocols before calling view_cb()
-    if (view_info.view >= 0) establish_protocol_versions (repl_proto);
-
     void*  app_req(0);
     size_t app_req_len(0);
 
@@ -1469,20 +1411,19 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
 
     if (view_info.view >= 0) // Primary configuration
     {
-        GU_DBUG_SYNC_WAIT("process_primary_configuration");
+        establish_protocol_versions (repl_proto);
 
         // we have to reset cert initial position here, SST does not contain
         // cert index yet (see #197).
-        // Also this must be done before releasing GCache buffers.
         cert_.assign_initial_position(group_seqno, trx_params_.version_);
-
-        if (STATE_SEQNO() > 0) service_thd_.release_seqno(STATE_SEQNO());
-        // make sure all gcache buffers are released
-
         // at this point there is no ongoing master or slave transactions
         // and no new requests to service thread should be possible
+
         service_thd_.flush();             // make sure service thd is idle
 
+        if (STATE_SEQNO() > 0) gcache_.seqno_release(STATE_SEQNO());
+        // make sure all gcache buffers are released
+
         // record state seqno, needed for IST on DONOR
         cc_seqno_ = group_seqno;
 
@@ -1500,7 +1441,6 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
             if (view_info.view == 1 || !app_wants_st)
             {
                 update_state_uuid (group_uuid);
-                gcache_.seqno_reset(to_gu_uuid(group_uuid), group_seqno);
                 apply_monitor_.set_initial_position(group_seqno);
                 if (co_mode_ != CommitOrder::BYPASS)
                     commit_monitor_.set_initial_position(group_seqno);
@@ -1532,7 +1472,7 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
                 }
             }
 
-            st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED, safe_to_bootstrap_);
+            st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED);
         }
 
         if (state_() == S_JOINING && sst_state_ != SST_NONE)
@@ -1543,7 +1483,7 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
              */
             try {
                 gcs_.join(sst_seqno_);
-                sst_state_ = SST_JOIN_SENT;
+                sst_state_ = SST_NONE;
             }
             catch (gu::Exception& e)
             {
@@ -1554,9 +1494,9 @@ galera::ReplicatorSMM::process_conf_change(void*                    recv_ctx,
     else
     {
         // Non-primary configuration
-        if (state_uuid_ != WSREP_UUID_UNDEFINED && next_state == S_CLOSING)
+        if (state_uuid_ != WSREP_UUID_UNDEFINED)
         {
-            st_.set (state_uuid_, STATE_SEQNO(), safe_to_bootstrap_);
+            st_.set (state_uuid_, STATE_SEQNO());
         }
 
         if (next_state != S_CONNECTED && next_state != S_CLOSING)
@@ -1599,7 +1539,6 @@ void galera::ReplicatorSMM::process_join(wsrep_seqno_t seqno_j,
     else
     {
         state_.shift_to(S_JOINED);
-        sst_state_ = SST_NONE;
     }
 
     local_monitor_.leave(lo);
@@ -1649,7 +1588,7 @@ wsrep_seqno_t galera::ReplicatorSMM::pause()
     }
 
     wsrep_seqno_t const ret(STATE_SEQNO());
-    st_.set(state_uuid_, ret, safe_to_bootstrap_);
+    st_.set(state_uuid_, ret);
 
     log_info << "Provider paused at " << state_uuid_ << ':' << ret
              << " (" << pause_seqno_ << ")";
@@ -1659,13 +1598,13 @@ wsrep_seqno_t galera::ReplicatorSMM::pause()
 
 void galera::ReplicatorSMM::resume()
 {
+    assert(pause_seqno_ != WSREP_SEQNO_UNDEFINED);
     if (pause_seqno_ == WSREP_SEQNO_UNDEFINED)
     {
-        log_warn << "tried to resume unpaused provider";
-        return;
+        gu_throw_error(EALREADY) << "tried to resume unpaused provider";
     }
 
-    st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED, safe_to_bootstrap_);
+    st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED);
     log_info << "resuming provider at " << pause_seqno_;
     LocalOrder lo(pause_seqno_);
     pause_seqno_ = WSREP_SEQNO_UNDEFINED;
@@ -1754,62 +1693,37 @@ wsrep_status_t galera::ReplicatorSMM::cert(TrxHandle* trx)
     }
 
     wsrep_status_t retval(WSREP_OK);
-    // IST should have drained the monitors, so STATE_SEQNO() should be current
     bool const applicable(trx->global_seqno() > STATE_SEQNO());
 
-    if (!applicable)
-    {
-        // this can happen after state transfer position has been submitted
-        // but not all actions preceding it have been processed.
-        //
-        // Cert index preload after SST:
-        // ----------------------------
-        // If the trx global seqno is in the half open range
-        // (cc_seqno_ , sst_seqno_], the write set was contained in the SST.
-        // In this case do the certification for trx to populate the index,
-        // but ignore the result. Always set state as S_MUST_ABORT and
-        // return WSREP_TRX_FAIL to make calling code to discard this trx.
-        if (last_st_type_ == ST_TYPE_SST &&
-            cc_seqno_ < trx->global_seqno() &&
-            trx->global_seqno() <= sst_seqno_)
-        {
-            (void)cert_.append_trx(trx);
-            trx->verify_checksum();
-            gcache_.seqno_assign (trx->action(),
-                                  trx->global_seqno(),
-                                  trx->depends_seqno());
-            cert_.set_trx_committed(trx);
-        }
-        else
-        {
-            gcache_.free(const_cast<void*>(trx->action()));
-        }
-        trx->set_state(TrxHandle::S_MUST_ABORT);
-        if (interrupted)
-            local_monitor_.self_cancel(lo);
-        else
-            local_monitor_.leave(lo);
-        return WSREP_TRX_FAIL;
-    }
-
     if (gu_likely (!interrupted))
     {
         switch (cert_.append_trx(trx))
         {
         case Certification::TEST_OK:
-            if (trx->state() == TrxHandle::S_CERTIFYING)
+            if (gu_likely(applicable))
             {
-                retval = WSREP_OK;
+                if (trx->state() == TrxHandle::S_CERTIFYING)
+                {
+                    retval = WSREP_OK;
+                }
+                else
+                {
+                    assert(trx->state() == TrxHandle::S_MUST_ABORT);
+                    trx->set_state(TrxHandle::S_MUST_REPLAY_AM);
+                    retval = WSREP_BF_ABORT;
+                }
             }
             else
             {
-                assert(trx->state() == TrxHandle::S_MUST_ABORT);
-                trx->set_state(TrxHandle::S_MUST_REPLAY_AM);
-                retval = WSREP_BF_ABORT;
+                // this can happen after SST position has been submitted
+                // but not all actions preceding SST initial position
+                // have been processed
+                trx->set_state(TrxHandle::S_MUST_ABORT);
+                retval = WSREP_TRX_FAIL;
             }
             break;
         case Certification::TEST_FAILED:
-            if (gu_unlikely(trx->is_toi())) // small sanity check
+            if (gu_unlikely(trx->is_toi() && applicable)) // small sanity check
             {
                 // may happen on configuration change
                 log_warn << "Certification failed for TO isolated action: "
@@ -1853,15 +1767,13 @@ wsrep_status_t galera::ReplicatorSMM::cert(TrxHandle* trx)
         }
     }
 
-    if (gu_unlikely(WSREP_TRX_FAIL == retval))
+    if (gu_unlikely(WSREP_TRX_FAIL == retval && applicable))
     {
         // applicable but failed certification: self-cancel monitors
         apply_monitor_.self_cancel(ao);
         if (co_mode_ != CommitOrder::BYPASS) commit_monitor_.self_cancel(co);
     }
 
-    assert(applicable);
-
     return retval;
 }
 
@@ -1922,15 +1834,12 @@ galera::ReplicatorSMM::update_state_uuid (const wsrep_uuid_t& uuid)
         *(const_cast<wsrep_uuid_t*>(&state_uuid_)) = uuid;
 
         std::ostringstream os; os << state_uuid_;
-        // Copy only non-nil terminated part of the source string
-        // and terminate the string explicitly to silence a warning
-        // generated by Wstringop-truncation
-        char* str(const_cast<char*>(state_uuid_str_));
-        strncpy(str, os.str().c_str(), sizeof(state_uuid_str_) - 1);
-        str[sizeof(state_uuid_str_) - 1] = '\0';
+
+        strncpy(const_cast<char*>(state_uuid_str_), os.str().c_str(),
+                sizeof(state_uuid_str_));
     }
 
-    st_.set(uuid, WSREP_SEQNO_UNDEFINED, safe_to_bootstrap_);
+    st_.set(uuid, WSREP_SEQNO_UNDEFINED);
 }
 
 void
