@@ -42,7 +42,6 @@ namespace gcomm
     std::string const Defaults::EvsDelayMargin          = "PT1S";
     std::string const Defaults::EvsDelayedKeepPeriod    = "PT30S";
     std::string const Defaults::EvsAutoEvict            = "0";
-    std::string const Defaults::EvsVersion              = "1";
     std::string const Defaults::PcAnnounceTimeout       = "PT3S";
     std::string const Defaults::PcChecksum              = "false";
     std::string const Defaults::PcIgnoreQuorum          = "false";
@@ -50,7 +49,7 @@ namespace gcomm
     std::string const Defaults::PcNpvo                  = "false";
     std::string const Defaults::PcVersion               = "0";
     std::string const Defaults::PcWaitPrim              = "true";
-    std::string const Defaults::PcWaitPrimTimeout       = "PT30S";
+    std::string const Defaults::PcWaitPrimTimeout       = "P30S";
     std::string const Defaults::PcWeight                = "1";
-    std::string const Defaults::PcRecovery              = "true";
+    std::string const Defaults::PcRecovery              = "1";
 }
