@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2024 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2013 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -12,18 +12,13 @@
 #include <stdint.h>
 #include <string.h>
 #include <galerautils.h>
-#include <gu_serialize.hpp>
-#include <gu_hexdump.h>
 
-#define GCS_STATE_MSG_VER 6
-#define GCS_STATE_MSG_NO_PROTO_DOWNGRADE_VER 6
+#define GCS_STATE_MSG_VER 4
 
 #define GCS_STATE_MSG_ACCESS
 #include "gcs_state_msg.hpp"
 #include "gcs_node.hpp"
 
-#include "gu_logger.hpp"
-
 gcs_state_msg_t*
 gcs_state_msg_create (const gu_uuid_t* state_uuid,
                       const gu_uuid_t* group_uuid,
@@ -31,36 +26,26 @@ gcs_state_msg_create (const gu_uuid_t* state_uuid,
                       gcs_seqno_t      prim_seqno,
                       gcs_seqno_t      received,
                       gcs_seqno_t      cached,
-                      gcs_seqno_t      last_applied,
-                      gcs_seqno_t      vote_seqno,
-                      int64_t          vote_res,
-                      uint8_t          vote_policy,
                       int              prim_joined,
                       gcs_node_state_t prim_state,
                       gcs_node_state_t current_state,
                       const char*      name,
                       const char*      inc_addr,
-                      int              gcs_proto_ver, /* max supported versions*/
+                      int              gcs_proto_ver,
                       int              repl_proto_ver,
                       int              appl_proto_ver,
-                      int              prim_gcs_ver, /* last prim versions*/
-                      int              prim_repl_ver,
-                      int              prim_appl_ver,
                       int              desync_count,
                       uint8_t          flags)
 {
 #define CHECK_PROTO_RANGE(LEVEL)                                        \
     if (LEVEL < (int)0 || LEVEL > (int)UINT8_MAX) {                     \
-        gu_error(#LEVEL " value %d is out of range [0, %d]", LEVEL,UINT8_MAX); \
+        gu_error ("#LEVEL value %d is out of range [0, %d]", LEVEL,UINT8_MAX); \
         return NULL;                                                    \
     }
 
     CHECK_PROTO_RANGE(gcs_proto_ver);
     CHECK_PROTO_RANGE(repl_proto_ver);
     CHECK_PROTO_RANGE(appl_proto_ver);
-    CHECK_PROTO_RANGE(prim_gcs_ver);
-    CHECK_PROTO_RANGE(prim_repl_ver);
-    CHECK_PROTO_RANGE(prim_appl_ver);
 
     size_t name_len = strlen(name) + 1;
     size_t addr_len = strlen(inc_addr) + 1;
@@ -76,19 +61,12 @@ gcs_state_msg_create (const gu_uuid_t* state_uuid,
         ret->prim_seqno    = prim_seqno;
         ret->received      = received;
         ret->cached        = cached;
-        ret->last_applied  = last_applied;
-        ret->vote_seqno    = vote_seqno;
-        ret->vote_res      = vote_res;
-        ret->vote_policy   = vote_policy;
         ret->prim_state    = prim_state;
         ret->current_state = current_state;
         ret->version       = GCS_STATE_MSG_VER;
         ret->gcs_proto_ver = gcs_proto_ver;
         ret->repl_proto_ver= repl_proto_ver;
         ret->appl_proto_ver= appl_proto_ver;
-        ret->prim_gcs_ver  = prim_gcs_ver;
-        ret->prim_repl_ver = prim_repl_ver;
-        ret->prim_appl_ver = prim_appl_ver;
         ret->desync_count  = desync_count;
         ret->name          = (char*)(ret + 1);
         ret->inc_addr      = ret->name + name_len;
@@ -135,17 +113,7 @@ gcs_state_msg_len (gcs_state_msg_t* state)
 // V3 stuff
         sizeof (int64_t)     +   // cached
 // V4 stuff
-        sizeof (int32_t)     +   // desync count
-// V5 stuff
-        sizeof (int64_t)     +   // last_applied
-        sizeof (int64_t)     +   // vote_seqno
-        sizeof (int64_t)     +   // vote_res
-        sizeof (uint8_t)     +   // vote_policy
-// V6 stuff
-        sizeof (int8_t)      +   // prim_gcs_ver
-        sizeof (int8_t)      +   // prim_repl_ver
-        sizeof (int8_t)      +   // prim_appl_ver
-        0
+        sizeof (int32_t)         // desync count
         );
 }
 
@@ -180,26 +148,16 @@ gcs_state_msg_len (gcs_state_msg_t* state)
     const char*      name           = (char*)(prim_seqno + 1);
 
 
+
 /* Serialize gcs_state_msg_t into buf */
 ssize_t
 gcs_state_msg_write (void* buf, const gcs_state_msg_t* state)
 {
     STATE_MSG_FIELDS_V0(buf);
     char*     inc_addr  = name + strlen (state->name) + 1;
-    uint8_t*  appl_proto_ver = (uint8_t*)(inc_addr + strlen(state->inc_addr) +1);
-// V3 stuff
+    uint8_t*  appl_proto_ver = (uint8_t*)(inc_addr + strlen(state->inc_addr) + 1);
     int64_t*  cached         = (int64_t*)(appl_proto_ver + 1);
-// V4 stuff
     int32_t*  desync_count   = (int32_t*)(cached + 1);
-// V5 stuff
-    int64_t* last_applied   = (int64_t*)(desync_count + 1);
-    int64_t* vote_seqno     = last_applied + 1;
-    int64_t* vote_res       = vote_seqno   + 1;
-    uint8_t* vote_policy    = (uint8_t*)(vote_res + 1);
-// V6 stuff
-    uint8_t*  prim_gcs_ver   = (uint8_t*)(vote_policy + 1);
-    uint8_t*  prim_repl_ver  = (uint8_t*)(prim_gcs_ver + 1);
-    uint8_t*  prim_appl_ver  = (uint8_t*)(prim_repl_ver + 1);
 
     *version        = GCS_STATE_MSG_VER;
     *flags          = state->flags;
@@ -213,31 +171,13 @@ gcs_state_msg_write (void* buf, const gcs_state_msg_t* state)
     *prim_uuid      = state->prim_uuid;
     *received       = htog64(state->received);
     *prim_seqno     = htog64(state->prim_seqno);
-
-    /* from this point alignment breaks */
     strcpy (name,     state->name);
     strcpy (inc_addr, state->inc_addr);
     *appl_proto_ver = state->appl_proto_ver; // in preparation for V1
+    *cached         = htog64(state->cached);
+    *desync_count   = htog32(state->desync_count);
 
-    gu::serialize8(state->cached, cached, 0);
-    gu::serialize4(state->desync_count, desync_count, 0);
-    gu::serialize8(state->last_applied, last_applied, 0);
-
-    gu::serialize8(state->vote_seqno, vote_seqno, 0); // 4.ee
-    gu::serialize8(state->vote_res, vote_res, 0);
-    gu::serialize1(state->vote_policy, vote_policy, 0);
-
-    *prim_gcs_ver    = state->prim_gcs_ver;
-    *prim_repl_ver   = state->prim_repl_ver;
-    *prim_appl_ver   = state->prim_appl_ver;
-
-    size_t const msg_len((uint8_t*)(prim_appl_ver + 1) - (uint8_t*)buf);
-#ifndef NDEBUG
-    char str[1024];
-    gu_hexdump(buf, msg_len, str, sizeof(str), true);
-    gu_debug("Serialized state message of size %zd\n%s", msg_len, str);
-#endif /* NDEBUG */
-    return msg_len;
+    return ((uint8_t*)(desync_count + 1) - (uint8_t*)buf);
 }
 
 /* De-serialize gcs_state_msg_t from buf */
@@ -246,12 +186,6 @@ gcs_state_msg_read (const void* const buf, ssize_t const buf_len)
 {
     assert (buf_len > 0);
 
-#ifndef NDEBUG
-    char str[1024];
-    gu_hexdump(buf, buf_len, str, sizeof(str), true);
-    gu_debug("Received state message of size %zd\n%s", buf_len, str);
-#endif /* NDEBUG*/
-
     /* beginning of the message is always version 0 */
     CONST_STATE_MSG_FIELDS_V0(buf);
     const char* inc_addr = name + strlen (name) + 1;
@@ -267,41 +201,14 @@ gcs_state_msg_read (const void* const buf, ssize_t const buf_len)
     int64_t* cached_ptr = (int64_t*)(appl_ptr + 1);
     if (*version >= 3) {
         assert(buf_len >= (uint8_t*)(cached_ptr + 1) - (uint8_t*)buf);
-        gu::unserialize8(cached_ptr, 0, cached);
+        cached = gtoh64(*cached_ptr);
     }
-// v4 stuff
+
     int32_t  desync_count = 0;
     int32_t* desync_count_ptr = (int32_t*)(cached_ptr + 1);
     if (*version >= 4) {
         assert(buf_len >= (uint8_t*)(desync_count_ptr + 1) - (uint8_t*)buf);
-        gu::unserialize4(desync_count_ptr, 0, desync_count);
-    }
-// v5 stuff
-    int64_t last_applied = 0;
-    int64_t vote_seqno   = 0;
-    int64_t vote_res     = 0;
-    uint8_t vote_policy  = GCS_VOTE_ZERO_WINS; // backward compatibility
-    int64_t* last_applied_ptr = (int64_t*)(desync_count_ptr + 1);
-    if (*version >= 5 && *gcs_proto_ver >= 2) {
-        assert(buf_len > (uint8_t*)(last_applied_ptr + 3) - (uint8_t*)buf);
-        gu::unserialize8(last_applied_ptr, 0, last_applied);
-
-        gu::unserialize8(last_applied_ptr + 1, 0, vote_seqno);
-        gu::unserialize8(last_applied_ptr + 2, 0, vote_res);
-        gu::unserialize1(last_applied_ptr + 3, 0, vote_policy);
-    }
-// v6 stuff
-    uint8_t prim_gcs_ver   = 0;
-    uint8_t* prim_gcs_ptr  = (uint8_t*)(last_applied_ptr + 3) + 1;
-    uint8_t prim_repl_ver  = 0;
-    uint8_t* prim_repl_ptr = (uint8_t*)(prim_gcs_ptr + 1);
-    uint8_t prim_appl_ver  = 0;
-    uint8_t* prim_appl_ptr = (uint8_t*)(prim_repl_ptr + 1);
-    if (*version >= 6) {
-        assert(buf_len >= (uint8_t*)(prim_appl_ptr + 1) - (uint8_t*)buf);
-        prim_gcs_ver    = *prim_gcs_ptr;
-        prim_repl_ver   = *prim_repl_ptr;
-        prim_appl_ver   = *prim_appl_ptr;
+        desync_count = gtoh32(*desync_count_ptr);
     }
 
     gcs_state_msg_t* ret = gcs_state_msg_create (
@@ -311,10 +218,6 @@ gcs_state_msg_read (const void* const buf, ssize_t const buf_len)
         gtoh64(*prim_seqno),
         gtoh64(*received),
         cached,
-        last_applied,
-        vote_seqno,
-        vote_res,
-        vote_policy,
         gtoh16(*prim_joined),
         (gcs_node_state_t)*prim_state,
         (gcs_node_state_t)*curr_state,
@@ -323,9 +226,6 @@ gcs_state_msg_read (const void* const buf, ssize_t const buf_len)
         *gcs_proto_ver,
         *repl_proto_ver,
         appl_proto_ver,
-        prim_gcs_ver,
-        prim_repl_ver,
-        prim_appl_ver,
         desync_count,
         *flags
         );
@@ -351,9 +251,6 @@ gcs_state_msg_snprintf (char* str, size_t size, const gcs_state_msg_t* state)
                      "\n\tPrim  seqno  : %lld"
                      "\n\tFirst seqno  : %lld"
                      "\n\tLast  seqno  : %lld"
-                     "\n\tCommit cut   : %lld"
-                     "\n\tLast vote    : %lld.%0llx"
-                     "\n\tVote policy  : %d"
                      "\n\tPrim JOINED  : %d"
                      "\n\tState UUID   : " GU_UUID_FORMAT
                      "\n\tGroup UUID   : " GU_UUID_FORMAT
@@ -370,9 +267,6 @@ gcs_state_msg_snprintf (char* str, size_t size, const gcs_state_msg_t* state)
                      (long long)state->prim_seqno,
                      (long long)state->cached,
                      (long long)state->received,
-                     (long long)state->last_applied,
-                     (long long)state->vote_seqno,(long long)state->vote_res,
-                     state->vote_policy,
                      state->prim_joined,
                      GU_UUID_ARGS(&state->state_uuid),
                      GU_UUID_ARGS(&state->group_uuid),
@@ -409,28 +303,6 @@ gcs_state_msg_cached (const gcs_state_msg_t* state)
     return state->cached;
 }
 
-/* Get last applied action seqno */
-gcs_seqno_t
-gcs_state_msg_last_applied (const gcs_state_msg_t* state)
-{
-    return state->last_applied;
-}
-
-/* Get last applied action vote */
-void
-gcs_state_msg_last_vote (const gcs_state_msg_t* state,
-                         gcs_seqno_t& seqno, int64_t& res)
-{
-    seqno = state->vote_seqno;
-    res   = state->vote_res;
-}
-
-uint8_t
-gcs_state_msg_vote_policy (const gcs_state_msg_t* state)
-{
-    return state->vote_policy;
-}
-
 /* Get current node state */
 gcs_node_state_t
 gcs_state_msg_current_state (const gcs_state_msg_t* state)
@@ -539,7 +411,7 @@ state_report_uuids (char* buf, size_t buf_len,
  * @retval (void*)-1 in case of fatal error */
 static const gcs_state_msg_t*
 state_quorum_inherit (const gcs_state_msg_t* states[],
-                      size_t                 states_num,
+                      long                   states_num,
                       gcs_state_quorum_t*    quorum)
 {
     /* They all must have the same group_uuid or otherwise quorum is impossible.
@@ -549,7 +421,7 @@ state_quorum_inherit (const gcs_state_msg_t* states[],
      * Of those with the status >= GCS_STATE_JOINED we choose the most
      * representative: with the highest act_seqno and prim_seqno.
      */
-    size_t i, j;
+    long i, j;
     const gcs_state_msg_t* rep = NULL;
 
     // find at least one JOINED/DONOR (donor was once joined)
@@ -568,11 +440,11 @@ state_quorum_inherit (const gcs_state_msg_t* states[],
             state_report_uuids (buf, buf_len, states, states_num,
                                 GCS_NODE_STATE_NON_PRIM);
 #ifdef GCS_CORE_TESTING
-            gu_info ("Quorum: No node with complete state:\n%s", buf);
+            gu_warn ("Quorum: No node with complete state:\n%s", buf);
 #else
             /* Print buf into stderr in order to message truncation
              * of application logger. */
-            gu_info ("Quorum: No node with complete state:");
+            gu_warn ("Quorum: No node with complete state:\n");
             fprintf(stderr, "%s\n", buf);
 #endif /* GCS_CORE_TESTING */
             gu_free (buf);
@@ -592,8 +464,7 @@ state_quorum_inherit (const gcs_state_msg_t* states[],
                 if (buf) {
                     state_report_uuids (buf, buf_len, states, states_num,
                                         GCS_NODE_STATE_DONOR);
-                    gu_fatal("Quorum impossible: conflicting group UUIDs:\n%s",
-                             buf);
+                    gu_fatal("Quorum impossible: conflicting group UUIDs:\n%s");
                     gu_free (buf);
                 }
                 else {
@@ -608,7 +479,6 @@ state_quorum_inherit (const gcs_state_msg_t* states[],
 
     quorum->act_id     = rep->received;
     quorum->conf_id    = rep->prim_seqno;
-    quorum->last_applied = rep->last_applied;
     quorum->group_uuid = rep->group_uuid;
     quorum->primary    = true;
 
@@ -697,7 +567,7 @@ state_quorum_remerge (const gcs_state_msg_t* const states[],
     struct candidate* candidates = GU_CALLOC(states_num, struct candidate);
 
     if (!candidates) {
-        gu_error ("Quorum: could not allocate %lu bytes for re-merge check.",
+        gu_error ("Quorum: could not allocate %zd bytes for re-merge check.",
                   states_num * sizeof(struct candidate));
         return NULL;
     }
@@ -804,7 +674,6 @@ state_quorum_remerge (const gcs_state_msg_t* const states[],
 
             quorum->act_id     = rep->received;
             quorum->conf_id    = rep->prim_seqno;
-            quorum->last_applied = rep->last_applied;
             quorum->group_uuid = rep->group_uuid;
             quorum->primary    = true;
         }
@@ -895,7 +764,6 @@ state_quorum_bootstrap (const gcs_state_msg_t* const states[],
 
             quorum->act_id     = rep->received;
             quorum->conf_id    = rep->prim_seqno;
-            quorum->last_applied = rep->last_applied;
             quorum->group_uuid = rep->group_uuid;
             quorum->primary    = true;
         }
@@ -914,13 +782,13 @@ state_quorum_bootstrap (const gcs_state_msg_t* const states[],
 /* Get quorum decision from state messages */
 long
 gcs_state_msg_get_quorum (const gcs_state_msg_t* states[],
-                          size_t                 states_num,
+                          long                   states_num,
                           gcs_state_quorum_t*    quorum)
 {
     assert (states_num > 0);
     assert (NULL != states);
 
-    size_t i;
+    long i;
     const gcs_state_msg_t*   rep = NULL;
 
     *quorum = GCS_QUORUM_NON_PRIMARY; // pessimistic assumption
@@ -956,10 +824,9 @@ gcs_state_msg_get_quorum (const gcs_state_msg_t* states[],
     INIT_PROTO_VER(gcs_proto_ver);
     INIT_PROTO_VER(repl_proto_ver);
     INIT_PROTO_VER(appl_proto_ver);
-#undef INIT_PROTO_VER
 
-    for (i = 0; i < states_num; i++)
-    {
+    for (i = 0; i < states_num; i++) {
+
 #define CHECK_MIN_PROTO_VER(LEVEL)                              \
         if (states[i]->LEVEL <  quorum->LEVEL) {                \
             quorum->LEVEL = states[i]->LEVEL;                   \
@@ -970,29 +837,6 @@ gcs_state_msg_get_quorum (const gcs_state_msg_t* states[],
             CHECK_MIN_PROTO_VER(repl_proto_ver);
             CHECK_MIN_PROTO_VER(appl_proto_ver);
 //        }
-#undef CHECK_MIN_PROTO_VER
-    }
-
-    if (quorum->version >= GCS_STATE_MSG_NO_PROTO_DOWNGRADE_VER)
-    {
-        // forbid protocol downgrade
-#define CHECK_MIN_PROTO_VER(LEVEL)                                      \
-        if (quorum->LEVEL##_proto_ver < rep->prim_##LEVEL##_ver) {      \
-            quorum->LEVEL##_proto_ver = rep->prim_##LEVEL##_ver;        \
-        }
-        CHECK_MIN_PROTO_VER(gcs);
-        CHECK_MIN_PROTO_VER(repl);
-        CHECK_MIN_PROTO_VER(appl);
-#undef CHECK_MIN_PROTO_VER
-    }
-
-    if (quorum->gcs_proto_ver < 1)
-    {
-        quorum->vote_policy = GCS_VOTE_ZERO_WINS;
-    }
-    else
-    {
-        quorum->vote_policy = rep->vote_policy;
     }
 
     if (quorum->version < 2) {;} // for future generations
